/*
   FNT - nodelist parser
   Nodelist/nodediff parser

   Copyright (C) 2009 Free Software Foundation, Inc.

   Written by:
   Slava Zanko <slavazanko@gmail.com>


   This file is part of the FIDO New Technology.

   FIDO New Technology is free software: you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as published 
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <stdio.h>
#include <string.h>

#include "../include/global.h"
#include "../include/types.h"
#include "../include/error.h"
#include "../include/sql.h"
#include "../include/fnar.h"

#include "main.h"

/*** global variables **************************************************/

/*** file scope macro definitions **************************************/

/*** file scope type declarations **************************************/

/*** file scope variables **********************************************/


static fnt_fnar_t *addr;

static char *phone;

/*** file scope functions **********************************************/

static void
ndl_fnar_clean (fnt_fnar_t * fnar)
{
    guint addrs_data;
    fnt_fnar_addr_t *addr;

    g_free (fnar->uri);
    fnar->uri = NULL;

    g_free (fnar->station_name);
    fnar->station_name = NULL;

    g_free (fnar->location);
    fnar->location = NULL;

    g_free (fnar->sysop);
    fnar->sysop = NULL;

    g_free (fnar->flags);
    fnar->flags = NULL;

    for (addrs_data = fnar->adresses->len; addrs_data > 0; addrs_data--)
    {
	addr = (fnt_fnar_addr_t *) g_ptr_array_index (fnar->adresses, 0);
	fnt_fnar_addr_del (fnar, addr);
    }


}
static int
ndl_parser_findRegion (void)
{
    int ret;
    fnt_fnar_t srch_addr;
    fnt_fnar_t *found_addr;
    GPtrArray *srch_results;

    memset (&srch_addr, 0, sizeof (fnt_fnar_t));

    srch_addr.zone = addr->zone;
    srch_addr.net = addr->net;
    srch_addr.node = addr->node;
    srch_addr.pvt = -1;
    srch_addr.down = -1;
    srch_addr.hold = -1;
    srch_addr.hub = -1;

    srch_results = fnt_fnar_search_local (&srch_addr);
    if (srch_results == NULL)
	return 0;

    found_addr = g_ptr_array_index (srch_results, 0);
    if (found_addr == NULL)
    {
	fnt_fnar_search_clearResults (&srch_results);
	return 0;
    }
    ret = found_addr->region;
    fnt_fnar_search_clearResults (&srch_results);
    return ret;
}

static GPtrArray *
ndl_parser_isNodeExists (void)
{
    fnt_fnar_t srch_addr;
    GPtrArray *srch_results;

    memset (&srch_addr, 0, sizeof (fnt_fnar_t));

    srch_addr.zone = addr->zone;
    srch_addr.net = addr->net;
    srch_addr.node = addr->node;
    srch_addr.point = addr->point;
    srch_addr.pvt = -1;
    srch_addr.down = -1;
    srch_addr.hold = -1;
    srch_addr.hub = -1;

    srch_results = fnt_fnar_search_local (&srch_addr);
    if (srch_results == NULL)
	return NULL;

    return srch_results;
}

static char *
ndl_parser_atStartOfString (char *pntr)
{
    char *tmp_pntr;

    if (!strncmp (pntr, "Zone,", 5))
    {
	pntr += 5;
	tmp_pntr = strchr (pntr, ',');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->zone = atoi (pntr);
	pntr = tmp_pntr + 1;
	addr->region = 0;
	addr->net = 0;
	addr->node = 0;
    }
    else if (!strncmp (pntr, "Region,", 7))
    {
	pntr += 7;
	tmp_pntr = strchr (pntr, ',');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->region = atoi (pntr);
	pntr = tmp_pntr + 1;

	addr->net = addr->region;
	addr->node = 0;
    }
    else if (!strncmp (pntr, "Host,", 5))
    {
	pntr += 5;
	tmp_pntr = strchr (pntr, ',');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->net = atoi (pntr);
	addr->node = 0;
	pntr = tmp_pntr + 1;
    }
    else if (!strncmp (pntr, "Pvt,", 4))
    {
	pntr += 4;
	tmp_pntr = strchr (pntr, ',');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->node = atoi (pntr);
	pntr = tmp_pntr + 1;

	addr->pvt = true;
    }
    else if (!strncmp (pntr, "Hold,", 5))
    {
	pntr += 5;
	tmp_pntr = strchr (pntr, ',');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->node = atoi (pntr);
	pntr = tmp_pntr + 1;

	addr->hold = true;
    }
    else if (!strncmp (pntr, "Down,", 5))
    {
	pntr += 5;
	tmp_pntr = strchr (pntr, ',');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->node = atoi (pntr);
	pntr = tmp_pntr + 1;

	addr->down = true;
    }
    else if (!strncmp (pntr, "Hub,", 4))
    {
	pntr += 4;
	tmp_pntr = strchr (pntr, ',');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->node = atoi (pntr);
	pntr = tmp_pntr + 1;

	addr->hub = true;
    }
    else if (!strncmp (pntr, "Boss,", 5))
    {
	pntr += 5;
	tmp_pntr = strchr (pntr, ':');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';

	addr->zone = atoi (pntr);
	pntr = tmp_pntr + 1;

	tmp_pntr = strchr (pntr, '/');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->net = atoi (pntr);
	pntr = tmp_pntr + 1;

	tmp_pntr = strchr (pntr, 0xd);
	if (tmp_pntr == NULL)
	    tmp_pntr = strchr (pntr, 0xa);

	if (tmp_pntr != NULL)
	    *tmp_pntr = '\0';
	addr->node = atoi (pntr);
	addr->region = ndl_parser_findRegion ();
	return NULL;
    }
    else if (!strncmp (pntr, "Point,", 4))
    {
	pntr += 6;
	tmp_pntr = strchr (pntr, ',');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->point = atoi (pntr);
	pntr = tmp_pntr + 1;
    }
    else if (!strncmp (pntr, ",", 1))
    {
	pntr += 1;
	tmp_pntr = strchr (pntr, ',');
	if (tmp_pntr == NULL)
	    return NULL;
	*tmp_pntr = '\0';
	addr->node = atoi (pntr);
	pntr = tmp_pntr + 1;
    }
    else
    {
	return NULL;
    }

    return pntr;
}

static char *
ndl_parser_getOneToken (char *pntr, char **token)
{
    char *tmp_pntr;
    tmp_pntr = strchr (pntr, ',');
    if (tmp_pntr == NULL)
	return NULL;
    *tmp_pntr = '\0';
    *token = g_strdup (pntr);
    pntr = tmp_pntr + 1;
    return pntr;
}

static char *
ndl_parser_getTokens (char *pntr)
{
    char *tmp_pntr;

    pntr = ndl_parser_getOneToken (pntr, &addr->station_name);
    if (!pntr)
	return NULL;

    pntr = ndl_parser_getOneToken (pntr, &addr->location);
    if (!pntr)
	return NULL;

    pntr = ndl_parser_getOneToken (pntr, &addr->sysop);
    if (!pntr)
	return NULL;

    pntr = ndl_parser_getOneToken (pntr, &phone);
    if (!pntr)
	return NULL;


    tmp_pntr = strchr (pntr, 0xd);
    if (tmp_pntr == NULL)
	tmp_pntr = strchr (pntr, 0xa);

    if (tmp_pntr != NULL)
	*tmp_pntr = '\0';

    addr->flags = g_strdup (pntr);
    return pntr;
}

static bool
ndl_parser_isFlagExists (const char *flags, const char *search_flag,
			 char **flag_option)
{
    char *copy_flags = g_strdup (flags);
    char *fnd_substr;
    char *delimeter;
    fnd_substr = strstr (copy_flags, search_flag);
    if (fnd_substr == NULL)
    {
	*flag_option = NULL;
	g_free (copy_flags);
	return false;
    }
    if (*(fnd_substr + strlen (search_flag)) == ':')
    {
	fnd_substr += strlen (search_flag) + 1;

	delimeter = strchr (fnd_substr, ',');
	if (delimeter == NULL)
	    delimeter = strchr (fnd_substr, 0xd);

	if (delimeter == NULL)
	    delimeter = strchr (fnd_substr, 0xa);

	if (delimeter != NULL)
	    *delimeter = '\0';
	*flag_option = g_strdup (fnd_substr);
    }
    g_free (copy_flags);
    return true;
}

static fnt_fnar_addrtype
ndl_parser_addr_getAddrType (const char *str_addr)
{
    if (g_regex_match_simple
	("^\\d+\\.\\d+\\.\\d+.\\d+(:\\d+)*$", str_addr, 0, 0))
	return FNT_FNAR_T_IPV4;

    if (g_regex_match_simple ("^[0-9a-fA-F:]+$", str_addr, 0, 0))
	return FNT_FNAR_T_IPV6;

    if (g_regex_match_simple ("\\.", str_addr, 0, 0))
	return FNT_FNAR_T_DNS;

    return FNT_FNAR_T_NONE;
}


static void
ndl_parser_add_addr_withInetAddr (const char *flags, const char *default_ip,
				  const char *flag,
				  fnt_fnar_addr_servicetype servicetype)
{
    char *flag_option = NULL;
    char *ip_port = NULL;
    fnt_fnar_addr_t fnar_addr;

    if (!ndl_parser_isFlagExists (flags, flag, &flag_option))
	return;

    fnar_addr.service_type = servicetype;

    if (flag_option == NULL)
    {
	fnar_addr.type = ndl_parser_addr_getAddrType (default_ip);
	fnar_addr.data = (char *) default_ip;
	if (fnar_addr.type != FNT_FNAR_T_NONE)
	    fnt_fnar_addr_add (addr, &fnar_addr);
	return;
    }

    if (g_regex_match_simple ("^\\d+$", flag_option, 0, 0))
    {
	fnar_addr.type = ndl_parser_addr_getAddrType (default_ip);

	if (fnar_addr.type == FNT_FNAR_T_IPV6)
	    ip_port = g_strdup_printf ("[%s]:%s", default_ip, flag_option);
	else
	    ip_port = g_strdup_printf ("%s:%s", default_ip, flag_option);
    }
    else
    {
	fnar_addr.type = ndl_parser_addr_getAddrType (flag_option);

	if (fnar_addr.type == FNT_FNAR_T_IPV6)
	    ip_port = g_strdup_printf ("[%s]", flag_option);
	else
	    ip_port = g_strdup (flag_option);
    }
    g_free (flag_option);
    fnar_addr.data = ip_port;
    if (fnar_addr.type != FNT_FNAR_T_NONE)
	fnt_fnar_addr_add (addr, &fnar_addr);
    g_free (ip_port);
}

static void
ndl_parser_add_addrs_email_one (const char *flags, const char *flag,
				const char *inetAddr,
				fnt_fnar_addr_servicetype servicetype)
{
    char *flag_option = NULL;
    fnt_fnar_addr_t fnar_addr;
    char *address;

    if (!ndl_parser_isFlagExists (flags, flag, &flag_option))
	return;

    if (flag_option)
	address = flag_option;
    else
	address = g_strdup (inetAddr);

    fnar_addr.type = FNT_FNAR_T_EMAIL;
    fnar_addr.service_type = servicetype;
    fnar_addr.data = address;
    fnt_fnar_addr_add (addr, &fnar_addr);
    g_free (address);
}


static void
ndl_parser_add_addrs_emails (const char *flags)
{
    char *flag_option = NULL;

    (void) ndl_parser_isFlagExists (flags, "EMA", &flag_option);

    if (!flag_option)
	(void) ndl_parser_isFlagExists (flags, "IEM", &flag_option);
    if (!flag_option)
	(void) ndl_parser_isFlagExists (flags, "UIC", &flag_option);
    if (!flag_option)
	(void) ndl_parser_isFlagExists (flags, "ITX", &flag_option);
    if (!flag_option)
	return;

    ndl_parser_add_addrs_email_one (flags, "EMA", flag_option,
				    FNT_FNAR_TSERVICE_EMAIL);
    ndl_parser_add_addrs_email_one (flags, "IEM", flag_option,
				    FNT_FNAR_TSERVICE_EMAIL);

    ndl_parser_add_addrs_email_one (flags, "UIC", flag_option,
				    FNT_FNAR_TSERVICE_UUENC);

    ndl_parser_add_addrs_email_one (flags, "ITX", flag_option,
				    FNT_FNAR_TSERVICE_TRANSX);

    ndl_parser_add_addrs_email_one (flags, "IMI", flag_option,
				    FNT_FNAR_TSERVICE_MIME);

    ndl_parser_add_addrs_email_one (flags, "ISE", flag_option,
				    FNT_FNAR_TSERVICE_SEAT);

    ndl_parser_add_addrs_email_one (flags, "EVY", flag_option,
				    FNT_FNAR_TSERVICE_VOYAGER);

    g_free (flag_option);
}

static void
ndl_parser_add_addrs (const char *flags)
{
    char *flag_option = NULL;
    char *ftndomain = NULL;
    char *inetAddr;
    fnt_fnar_addr_t fnar_addr;

    if (strcasecmp (phone, "-Unpublished-")
	&& strcasecmp (phone, "-Unlisted-"))
    {
	fnar_addr.type = FNT_FNAR_T_PHONE;
	fnar_addr.service_type = FNT_FNAR_TSERVICE_NONE;
	fnar_addr.data = phone;
	fnt_fnar_addr_add (addr, &fnar_addr);
    }
    fnt_fnar_convert_ftn_to_uri (addr);
    if (ndl_parser_isFlagExists (flags, "DOM", &flag_option))
    {
	if (flag_option)
	{
	    ftndomain =
		g_strdup_printf ("%s.fidonet.%s", addr->uri, flag_option);
	    g_free (flag_option);
	}
	else
	{
	    inetAddr = addr->station_name;
	}
    }
    else
    {
	inetAddr = addr->station_name;
    }

    /* present INA or IP flag */

    if (ndl_parser_isFlagExists (flags, "IP", &flag_option) ||
	ndl_parser_isFlagExists (flags, "INA", &flag_option))
    {
	if (flag_option != NULL)
	    inetAddr = flag_option;
    }

    /* no present INA or IP flag, but may be present other flags */
    ndl_parser_add_addr_withInetAddr (flags, inetAddr, "IBN",
				      FNT_FNAR_TSERVICE_BINKP);
    ndl_parser_add_addr_withInetAddr (flags, inetAddr, "IFC",
				      FNT_FNAR_TSERVICE_IFCICO);
    ndl_parser_add_addr_withInetAddr (flags, inetAddr, "IFT",
				      FNT_FNAR_TSERVICE_FTP);
    ndl_parser_add_addr_withInetAddr (flags, inetAddr, "ITN",
				      FNT_FNAR_TSERVICE_TELNET);
    ndl_parser_add_addr_withInetAddr (flags, inetAddr, "IVM",
				      FNT_FNAR_TSERVICE_VMODEM);
    ndl_parser_add_addr_withInetAddr (flags, inetAddr, "FNT",
				      FNT_FNAR_TSERVICE_FNT);

    g_free (flag_option);
    g_free (ftndomain);

    ndl_parser_add_addrs_emails (flags);

}

/*** public functions **************************************************/

void
ndl_parser_run (int file)
{
    FILE *fd;
    char buffer[1024];
    char *pntr, *tmp_pntr;
    GPtrArray *srch_results;

    fd = fdopen (file, "r");
    if (!fd)
    {
	fnt_error_push (0, 0,
			_("ERROR: Unable to operate with file descriptor\n"));
	ndl_exit (2);
    }
    addr = fnt_fnar_new ();

    memset (buffer, 0, 1024);

    while (fgets (buffer, 1023, fd))
    {
	for (pntr = buffer;
	     pntr - buffer < 1024 && (*pntr == ' ' || *pntr == '\t'); pntr++);
	if (*pntr == ';')
	    continue;

	if (addr == NULL)
	    return;

	addr->hold = false;
	addr->down = false;
	addr->pvt = false;
	addr->hub = false;
	addr->point = 0;

	pntr = ndl_parser_atStartOfString (pntr);
	if (pntr == NULL)
	{
	    ndl_fnar_clean (addr);
	    continue;
	}

	pntr = ndl_parser_getTokens (pntr);
	if (pntr == NULL)
	{
	    ndl_fnar_clean (addr);
	    continue;
	}
	ndl_parser_add_addrs (pntr);

	srch_results = ndl_parser_isNodeExists ();
	if (srch_results != NULL)
	{
	    // edit
	    fnt_fnar_t *tmp;

	    tmp = g_ptr_array_index (srch_results, 0);
	    if (!fnt_fnar_change (tmp->db_id, addr, true))
	    {
		printf (_("Error:\n%s"), fnt_error_getStackAsString ());
	    }

	    fnt_fnar_search_clearResults (&srch_results);
	}
	else
	{
	    // add

	    if (!fnt_fnar_add (addr, true))
	    {
		printf (_("Error:\n%s"), fnt_error_getStackAsString ());
	    }

	}
	ndl_fnar_clean (addr);
    }
    fnt_fnar_free (&addr);
    fclose (fd);
}
